#include "mcdp.h"

/**
 * Copyright (C) 2001-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * linux provides two interfaces:
 * - we can use both of them, but have to test what interface the devicefile
 *   supports!
 *
 * todo:
 * - add support for multiple cdroms
 * - ...
 */
#if defined(WANT_LINUX_SCSI_GENERIC) \
 || defined(WANT_LINUX_SEND_PACKET)

/* from /usr/src/linux/include/linux/major.h */
#define SCSI_CDROM_MAJOR	11
#define SCSI_GENERIC_MAJOR	21
#define SCSI_CHANGER_MAJOR	86

char *type_pc="linux packet command";
char *type_sg="linux scsi generic";

#include <limits.h>
#include <stdlib.h>
#include <libgen.h>

static int cd_init_resolve_symlink(char *dev, struct stat *st)
{
	int mydir=open(".",O_RDONLY); /* save directory */
	char *d=strdup(dev), *d2;
	char *f=strdup(dev), *f2;
	char real[PATH_MAX+1];

	if (!d) die_nomem();
	if (!f) die_nomem();

	/* should we add here some error checking ? */
	d2=dirname(d);
	f2=basename(f);
	chdir(d2);
	readlink(f2, real, PATH_MAX);
	if (lstat(real, st) == -1) {
		free(d); free(f);
		fchdir(mydir);
		close(mydir);
		return -1;
	}

	if (S_ISLNK(st->st_mode)) {
		/* try it recursivly */
		cd_init_resolve_symlink(real, st);
	}

	free(d); free(f);
	fchdir(mydir);
	close(mydir);

	return 0;
}

static int cd_init_tryit(int *fd, char *dev)
{
	struct stat st;
	int r;

	if (!dev) return -1;

	r = open(dev, O_RDWR|O_NONBLOCK);
	if (r == -1) {
		return -1;
	}

	if (lstat(dev, &st) == -1) {
		return -1; /* can't happen, because we opened it already! */
	}

	if (S_ISLNK(st.st_mode)) {
		if (cd_init_resolve_symlink(dev, &st) == -1) {
			die_msg(2, lang_msg(MSG_ERROR_DEV_LSTAT));
		}
	}

	if (!S_ISCHR(st.st_mode) && !S_ISBLK(st.st_mode)) {
		die_msg(2, lang_msg(MSG_ERROR_DEV_FILE));
	}

	/* device seems okay, fd can be set */
	*fd = r;

	switch (st.st_rdev>>8) {
		case SCSI_CDROM_MAJOR:
		case SCSI_GENERIC_MAJOR:
		case SCSI_CHANGER_MAJOR:
			/* LINUX_SCSI_GENERIC interface */
			cd->driver=type_sg;
			cd->cmd=cmd_lx_sg;
			break;
		default:
			/* LINUX_SEND_PACKET interface */
			cd->driver=type_pc;
			cd->cmd=cmd_lx_sp;
			break;
	}


	/**
	 * we use our scmd_xxx() functions
	 */
	cd->cd_stop=scmd_stop;
	cd->cd_eject=scmd_eject;
	cd->cd_close=scmd_close;
	cd->cd_pauseresume=scmd_pauseresume;
	cd->cd_inquiry=scmd_inquiry;
	cd->cd_initdisc=scmd_initdisc;
	cd->cd_play=scmd_play;
	cd->cd_getstatus=cd_getstatus;
	cd->cd_seek=cd_seek;
	cd->cd_volume=cd_volume;

	byte_zero(cd->vendor, sizeof(cd->vendor));
	byte_zero(cd->model,  sizeof(cd->model));
	byte_zero(cd->rev,    sizeof(cd->rev));

	stralloc_init(&cdb.cmd);
	stralloc_init(&cdb.buf);
	stralloc_init(&cdb.sense);

	stralloc_init(&cd->year);
	stralloc_init(&cd->genre);
	stralloc_init(&cd->cddbfile);
	stralloc_init(&cd->playorder);

	if(!stralloc_copys(&cd->year, "")) die_nomem();
	if(!stralloc_copys(&cd->genre, "")) die_nomem();
	if(!stralloc_copys(&cd->cddbfile, "")) die_nomem();
	if(!stralloc_copys(&cd->playorder, "")) die_nomem();

	cd->cd_inquiry();

	return 0;
}

/* setup the cdrom device @ linux systems */
void dev_linux(void)
{
	cd = array_allocate(&array_cds, sizeof(struct mcdp), 0);
	if (!cd) die_nomem();
	byte_zero(cd, sizeof(struct mcdp));

	/* 1. check argv[1] */
	if (mcdp_argc == 2) {
		cd->dev=mcdp_argv[1];
		if (cd_init_tryit(&cd->fd, cd->dev) == -1) {
			die_dev(cd->dev);
		}
	} else {
		cd->dev=mcdp_env_devcdrom;
		if (cd_init_tryit(&cd->fd, cd->dev) == -1) {
			die_dev(cd->dev);
		}
	}
	cd = array_start(&array_cds);

	return;
}

void cd_getstatus(void)
{
	static unsigned int last_lba_state=0;
	struct cdrom_volctrl cdv;
	struct cdrom_subchnl cds;
	int r;

	if (cd->ds != DS_OKAY) {
		r=ioctl(cd->fd, CDROM_DRIVE_STATUS, 0);
		if (r != CDS_DISC_OK) {
			cd->ds=DS_NODISC;
			return;
		}

		r=ioctl(cd->fd, CDROM_DISC_STATUS, 0);
		switch (r) {
			case CDS_AUDIO:
			case CDS_MIXED:
				cd->ds=DS_OKAY;
				break;
			default:
				cd->ds=DS_NODISC;
				return;
		}
	}

	/* we read in msf format */
	cds.cdsc_format = CDROM_LBA;
	if (ioctl(cd->fd, CDROMSUBCHNL, &cds) == -1) {
		cd->ds = DS_BUSY;
		return;
	}
	cd->lba=cds.cdsc_absaddr.lba+150;
	cd->rel=cds.cdsc_reladdr.lba+150;
	
	if (last_lba_state != cd->lba) {
		set(cd->updates, U_TIME);
		last_lba_state=cd->lba;
	}

	/* audiostatus changed */
	switch (cds.cdsc_audiostatus) {
		case CDROM_AUDIO_PLAY:
			if (cd->as != AS_PLAY) {
				cd->as = AS_PLAY;
				set(cd->updates, U_STATUS);
			}
			break;
		case CDROM_AUDIO_PAUSED:
			if (cd->as != AS_PAUSE) {
				cd->as = AS_PAUSE;
				set(cd->updates, U_STATUS);
			}
			break;
		case CDROM_AUDIO_COMPLETED:
			if (cd->as != AS_DONE) {
				cd->as = AS_DONE;
				set(cd->updates, U_STATUS);
			}
			break;
		case CDROM_AUDIO_ERROR:
			if (cd->as != AS_ERROR) {
				cd->as = AS_ERROR;
				set(cd->updates, U_STATUS);
			}
			break;
		case CDROM_AUDIO_NO_STATUS:
		case CDROM_AUDIO_INVALID:
			if (cd->as != AS_NOSTATUS) {
				cd->as = AS_NOSTATUS;
				set(cd->updates, U_STATUS);
			}
			break;
	}

	/* volume(0..3) changed */
	if (ioctl(cd->fd, CDROMVOLREAD, &cdv) == -1) {
		cd->ds = DS_BUSY;
		return;
	}

	if (cdv.channel0 != cd->volume) {
		cd->volume=cdv.channel0;
		set(cd->updates, U_VOLUME);
	}

	return;
}

void cd_volume(int volume)
{
	struct cdrom_volctrl vol;

	vol.channel0=vol.channel1=vol.channel2=vol.channel3=volume;
	(void)ioctl(cd->fd, CDROMVOLCTRL, &vol);
	set(cd->updates, U_VOLUME);
}

void cd_seek(int secs)
{
	int i=secs;

	secs=i;

	return;
}

#endif /* WANT_LINUX_... */
